@@ -1,10 +1,17 @@
+import glob
+import logging
 import os
-from dataclasses import dataclass
+from dataclasses import dataclass, field
+from pathlib import Path
 
 import tomlkit
 from tomlkit import TOMLDocument
 from tomlkit.container import Container
 
+from .util import console
+
+logger = logging.getLogger(__name__)
+
 
 @dataclass
 class Config:
@@ -13,18 +20,13 @@ class Config:
 
     def get_env(self, key: str, default: str | None = None) -> str | None:
         """Gets an enviromnent variable, checks the config file if it's not set in the environment."""
-        try:
-            return self.get_env_required(key)
-        except KeyError:
-            return default
+        return os.environ.get(key) or self.env.get(key) or default
 
     def get_env_required(self, key: str) -> str:
         """Gets an enviromnent variable, checks the config file if it's not set in the environment."""
-        if key in os.environ:
-            return os.environ[key]
-        elif key in self.env:
-            return self.env[key]
-        raise KeyError(
+        if val := os.environ.get(key) or self.env.get(key):
+            return val
+        raise KeyError(  # pragma: no cover
             f"Environment variable {key} not set in env or config, see README."
         )
 
@@ -35,6 +37,13 @@ class Config:
         }
 
 
+@dataclass
+class ProjectConfig:
+    """Project-level configuration, such as which files to include in the context by default."""
+
+    files: list[str] = field(default_factory=list)
+
+
 ABOUT_ACTIVITYWATCH = """ActivityWatch is a free and open-source automated time-tracker that helps you track how you spend your time on your devices."""
 ABOUT_GPTME = "gptme is a CLI to interact with large language models in a Chat-style interface, enabling the assistant to execute commands and code on the local machine, letting them assist in all kinds of development and terminal-based work."
 
@@ -81,7 +90,7 @@ def _load_config() -> tomlkit.TOMLDocument:
         os.makedirs(os.path.dirname(config_path), exist_ok=True)
         with open(config_path, "w") as config_file:
             tomlkit.dump(default_config.dict(), config_file)
-        print(f"Created config file at {config_path}")
+        console.log(f"Created config file at {config_path}")
 
     # Now you can read the settings from the config file like this:
     with open(config_path) as config_file:
@@ -89,7 +98,7 @@ def _load_config() -> tomlkit.TOMLDocument:
     return doc
 
 
-def set_config_value(key: str, value: str) -> None:
+def set_config_value(key: str, value: str) -> None:  # pragma: no cover
     doc: TOMLDocument | Container = _load_config()
 
     # Set the value
@@ -108,6 +117,36 @@ def set_config_value(key: str, value: str) -> None:
     _config = load_config()
 
 
+def get_workspace_prompt(workspace: str) -> str:
+    project_config_paths = [
+        p
+        for p in (
+            Path(workspace) / "gptme.toml",
+            Path(workspace) / ".github" / "gptme.toml",
+        )
+        if p.exists()
+    ]
+    if project_config_paths:
+        project_config_path = project_config_paths[0]
+        console.log(f"Using project configuration at {project_config_path}")
+        # load project config
+        with open(project_config_path) as f:
+            project_config = tomlkit.load(f)
+            project = ProjectConfig(**project_config)  # type: ignore
+            # expand with glob
+            files = [p for file in project.files for p in glob.glob(file)]
+            for file in files:
+                if not Path(file).exists():
+                    logger.error(
+                        f"File {file} specified in project config does not exist"
+                    )
+                    exit(1)
+        return "\n\nSelected project files, read more with cat:\n" + "\n\n".join(
+            [f"```{Path(file).name}\n{Path(file).read_text()}\n```" for file in files]
+        )
+    return ""
+
+
 if __name__ == "__main__":
     config = get_config()
     print(config)
