@@ -1,5 +1,12 @@
-# prefix for commands, e.g. /help
-CMDFIX = "/"
+"""
+Constants
+"""
+
+# Optimized for code
+# Discussion here: https://community.openai.com/t/cheat-sheet-mastering-temperature-and-top-p-in-chatgpt-api-a-few-tips-and-tricks-on-controlling-the-creativity-deterministic-output-of-prompt-responses/172683
+# TODO: make these configurable
+TEMPERATURE = 0
+TOP_P = 0.1
 
 # separator for multiple rounds of prompts on the command line
 # demarcates the end of the user's prompt, and start of the assistant's response
