@@ -16,8 +16,8 @@ from collections.abc import Generator
 import bashlex
 
 from ..message import Message
-from ..util import ask_execute, print_preview, transform_examples_to_chat_directives
-from .base import ToolSpec
+from ..util import ask_execute, get_tokenizer, print_preview
+from .base import ToolSpec, ToolUse
 
 logger = logging.getLogger(__name__)
 
@@ -56,13 +56,11 @@ These programs are available, among others:
 {shell_programs_str}
 """.strip()
 
-examples = """
+examples = f"""
 
 User: list the current directory
 Assistant: To list the files in the current directory, use `ls`:
-```bash
-ls
-```
+{ToolUse("bash", [], "ls").to_output()}
 System: Ran command: `ls`
 ```stdout
 file1.txt
@@ -72,9 +70,7 @@ file2.txt
 #### The assistant can learn context by exploring the filesystem
 User: learn about the project
 Assistant: Lets start by checking the files
-```bash
-git ls-files
-```
+{ToolUse("bash", [], "git ls-files").to_output()}
 System:
 ```output
 README.md
@@ -89,14 +85,27 @@ System:
 (contents of README.md)
 ```
 Assistant: Now we check main.py
-```bash
-cat main.py
-```
+{ToolUse("bash", [], "cat main.py").to_output()}
 System:
 ```output
 (contents of main.py)
 ```
 Assistant: The project is...
+
+
+#### Create vue project
+User: Create a new vue project with typescript and pinia named fancy-project
+Assistant: Sure! Let's create a new vue project with TypeScript and Pinia named fancy-project:
+{ToolUse("bash", [], "npm init vue@latest fancy-project --yes -- --typescript --pinia").to_output()}
+System:
+```output
+> npx
+> create-vue
+
+Vue.js - The Progressive JavaScript Framework
+
+Scaffolding project in ./fancy-project...
+```
 """.strip()
 
 
@@ -218,7 +227,7 @@ class ShellSession:
         self._init()
 
 
-_shell = None
+_shell: ShellSession | None = None
 
 
 def get_shell() -> ShellSession:
@@ -253,9 +262,13 @@ def execute_shell(
         print()
 
     if not ask or confirm:
-        returncode, stdout, stderr = shell.run(cmd)
-        stdout = _shorten_stdout(stdout.strip())
-        stderr = _shorten_stdout(stderr.strip())
+        try:
+            returncode, stdout, stderr = shell.run(cmd)
+        except Exception as e:
+            yield Message("system", f"Error: {e}")
+            return
+        stdout = _shorten_stdout(stdout.strip(), pre_tokens=2000, post_tokens=8000)
+        stderr = _shorten_stdout(stderr.strip(), pre_tokens=2000, post_tokens=2000)
 
         msg = _format_block_smart("Ran command", cmd, lang="bash") + "\n\n"
         if stdout:
@@ -282,10 +295,11 @@ def _shorten_stdout(
     stdout: str,
     pre_lines=None,
     post_lines=None,
+    pre_tokens=None,
+    post_tokens=None,
     strip_dates=False,
     strip_common_prefix_lines=0,
 ) -> str:
-    """Shortens stdout to 1000 tokens."""
     lines = stdout.split("\n")
 
     # NOTE: This can cause issues when, for example, reading a CSV with dates in the first column
@@ -320,6 +334,18 @@ def _shorten_stdout(
             + lines[-post_lines:]
         )
 
+    # check that if pre_tokens is set, so is post_tokens, and vice versa
+    assert (pre_tokens is None) == (post_tokens is None)
+    if pre_tokens is not None and post_tokens is not None:
+        tokenizer = get_tokenizer("gpt-4")  # TODO: use sane default
+        tokens = tokenizer.encode(stdout)
+        if len(tokens) > pre_tokens + post_tokens:
+            lines = (
+                [tokenizer.decode(tokens[:pre_tokens])]
+                + ["... (truncated output) ..."]
+                + [tokenizer.decode(tokens[-post_tokens:])]
+            )
+
     return "\n".join(lines)
 
 
@@ -343,24 +369,22 @@ def split_commands(script: str) -> list[str]:
                     command_parts.append(script[start:end])
                 command = " ".join(command_parts)
                 commands.append(command)
-        elif part.kind == "function":
-            commands.append(script[part.pos[0] : part.pos[1]])
-        elif part.kind == "pipeline":
+        elif part.kind in ["function", "pipeline", "list"]:
             commands.append(script[part.pos[0] : part.pos[1]])
         else:
-            logger.warning(f"Unknown shell script part of kind '{part.kind}', skipping")
+            logger.warning(
+                f"Unknown shell script part of kind '{part.kind}', hoping this works"
+            )
+            commands.append(script[part.pos[0] : part.pos[1]])
     return commands
 
 
-__doc__ += transform_examples_to_chat_directives(examples)
-
-
 tool = ToolSpec(
     name="shell",
     desc="Executes shell commands.",
     instructions=instructions,
     examples=examples,
-    init=get_shell,
     execute=execute_shell,
-    block_types=["bash", "sh"],
+    block_types=["bash", "sh", "shell"],
 )
+__doc__ = tool.get_doc(__doc__)
